import copy
import datetime
import json
import logging
import os
import signal
import time
from eliot import start_action
from kubernetes import client, config
from jupyterhubutils.scanrepo import ScanRepo
from threading import Thread
from ..utils import make_logger


class Prepuller(object):
    '''Class for generating and reaping the Pods for the prepuller.
    '''

    def __init__(self, args=None):
        self.logger = make_logger()
        self.repo = None
        if not args:
            raise ValueError("args must be set!")
        self.args = args
        self.debug = False
        if args.debug:
            self.debug = True
        if self.debug:
            self.logger.setLevel(logging.DEBUG)
            self.logger.debug("Debug logging enabled.")
        namespace = self.args.namespace
        if not namespace:
            self.logger.warning("Using kubernetes namespace 'default'")
            namespace = "default"
        self.namespace = namespace
        try:
            config.load_incluster_config()
        except config.ConfigException:
            config.load_kube_config()
        self.client = client.CoreV1Api()
        self.images = []
        self.nodes = []
        self.pod_specs = {}
        self.created_pods = []

        self.logger.debug("Arguments: %s" % str(args))
        self.command = self.args.command
        self.cachefile = self.args.cachefile
        if self.args.list:
            for image in self.args.list:
                # Make fully-qualified image name
                colons = image.count(':')
                if colons == 0:
                    image = image + ":latest"
                slashes = image.count('/')
                if slashes == 0:
                    image = "library/" + image
            self.images.append(image)
        # Cheap way to deduplicate lists
        self.images = list(set(self.images))
        if self.images:
            self.images.sort()
        # Not portable to non-Unixy systems.
        if self.args.timeout >= 0:
            self.logger.debug("Setting timeout to %d s." % self.args.timeout)
            signal.signal(signal.SIGALRM, self._timeout_handler)
            signal.alarm(self.args.timeout)

    def _timeout_handler(self, signum, frame):
        with start_action(action_type="_timeout_handler"):
            self.logger.error(
                "Did not complete in %d s.  Terminating." % self.args.timeout)
            self._destroy_pods(selective=False)
            raise RuntimeError("Timed out")

    def _destroy_pods(self, selective=False):
        '''Get a pod list and delete any that are still running if
        selective is False, or any in state "Succeeded" or "Failed"
        if selective is True.
        '''
        with start_action(action_type="_destroy_pods"):
            self.logger.debug("Looking for pods to delete.")
            cleanup = self._get_deletion_list(selective=selective)
            for podname in cleanup:
                self.delete_pod(podname)

    def _get_deletion_list(self, selective=True):
        with start_action(action_type="_get_deletion_list"):
            cleanup = []
            speclist = []
            v1 = self.client
            for x in self.pod_specs:
                speclist.extend(self.pod_specs[x])
            specnames = [self._derive_pod_name(x) for x in speclist]
            podlist = v1.list_namespaced_pod(self.namespace)
            for pod in podlist.items:
                podname = pod.metadata.name
                phase = pod.status.phase
                if podname in specnames:
                    if selective:
                        if (phase != "Succeeded" and phase != "Failed"):
                            self.logger.debug("Pod '%s' %s; not cleaning."
                                              % (podname, phase))
                            continue
                    self.logger.debug(
                        "Pod '%s' %s; adding to cleanup." % (podname, phase))
                    cleanup.append(podname)
            return cleanup

    def update_images_from_repo(self):
        '''Scan the repo looking for images.
        '''
        with start_action(action_type="update_images_from_repo"):
            if not self.repo:
                self.repo = ScanRepo(host=self.args.repo,
                                     path=self.args.path,
                                     owner=self.args.owner,
                                     name=self.args.name,
                                     dailies=self.args.dailies,
                                     weeklies=self.args.weeklies,
                                     releases=self.args.releases,
                                     experimentals=self.args.experimentals,
                                     recommended=self.args.recommended,
                                     json=True, insecure=self.args.insecure,
                                     sort_field=self.args.sort,
                                     cachefile=self.cachefile,
                                     debug=self.args.debug)
            if not self.args.no_scan:
                if self.args.repo:
                    self.logger.debug("Scanning images: '%s' " %
                                      self.args.repo)
                else:
                    self.logger.debug("Scanning Docker repo for images")
                self.repo.scan()
                if self.repo.debug:
                    repocopy = copy.deepcopy(self.repo.data)
                    for itype in repocopy:
                        for item in repocopy[itype]:
                            uddt = item["updated"]
                            if type(uddt) is datetime.datetime:
                                uddt = uddt.isoformat()
                                item["updated"] = uddt
                    self.logger.debug("Scan Data: " +
                                      json.dumps(repocopy,
                                                 sort_keys=True,
                                                 indent=4))
                scan_imgs = []
                sections = []
                if self.args.recommended:
                    sections.extend(["recommended"])
                sections.extend(["experimental", "daily", "weekly", "release"])
                for section in sections:
                    if section not in self.repo.data:
                        continue
                    for entry in self.repo.data[section]:
                        exhost = ''
                        if self.args.repo:
                            exhost = self.args.repo
                            if self.args.port:
                                exhost += ":" + self.args.port
                        if exhost == "hub.docker.com":
                            # We could change it to "index.docker.io" but
                            #  leaving it empty works too.
                            exhost = ""
                            if exhost and exhost[-1] != "/":
                                exhost += "/"
                            scan_imgs.append(exhost + self.args.owner + "/" +
                                             self.args.name + ":" +
                                             entry["name"])
                current_imgs = [x for x in self.images]
                # Dedupe by running the list through a set.
                current_imgs.extend(scan_imgs)
                current_imgs = list(set(current_imgs))
                if current_imgs:
                    current_imgs.sort()
                self.images = current_imgs

    def build_nodelist(self):
        '''Make a list of all schedulable nodes, respecting RESTRICT_*
        environment variables.
        '''
        with start_action(action_type="build_nodelist"):
            v1 = self.client
            logger = self.logger
            logger.debug("Getting schedulable node list.")
            v1nodelist = v1.list_node()
            nodes = []
            for thing in v1nodelist.items:
                spec = thing.spec
                if spec.unschedulable:
                    continue
                if spec.taints:
                    taints = [x.effect for x in spec.taints]
                    if "NoSchedule" in taints:
                        continue
                if self.reject_by_label(thing):
                    continue
                nodes.append(thing.metadata.name)
            logger.debug("Schedulable list: %s" % str(nodes))
            self.nodes = nodes

    def reject_by_label(self, node):
        '''If node labels are set to restrict Lab spawn, reject nodes that
        are not suitable for Lab/Dask.
        '''
        with start_action(action_type="reject_by_label"):
            logger = self.logger
            logger.debug("Checking for node labels.")
            if not os.getenv("RESTRICT_LAB_NODES"):
                logger.debug("Lab nodes are not restricted.")
                return False
            labels = node.metadata.labels
            name = node.metadata.name or 'Node Name Unknown'
            if not labels:
                logger.debug("Nodes are not labelled.")
                return False
            lab_ok = labels.get("jupyterlab")
            if lab_ok and lab_ok == "ok":
                logger.debug("Node '%s' is allowed for Lab usage." % name)
                return False
            if not os.getenv("ALLOW_DASK_SPAWN"):
                logger.debug("Lab spawn not allowed for node '%s'." % name)
                return True
            if not os.getenv("RESTRICT_DASK_NODES"):
                logger.debug("Dask allowed and unrestricted.")
                return False
            dask_ok = labels.get("dask")
            if dask_ok and dask_ok == "ok":
                logger.debug("Node '%s' is allowed for Dask usage." % name)
                return False
            logger.debug("Lab/Dask spawn not allowed for node '%s'" % name)
            return True

    def build_pod_specs(self):
        '''Build a dict of Pod specs by node, each node having a list of
        specs.
        '''
        with start_action(action_type="build_pod_specs"):
            specs = {}
            for node in self.nodes:
                specs[node] = []
                for img in self.images:
                    specs[node].append(self._build_pod_spec(img, node))
            self.pod_specs = specs
            self.logger.debug("Specs: %s" % str(self.pod_specs))

    def _build_pod_spec(self, img, node):
        with start_action(action_type="_build_pod_spec"):
            spec = client.V1PodSpec(
                containers=[
                    client.V1Container(
                        command=self.command,
                        image=img,
                        image_pull_policy="Always",
                        name=self._podname_from_image(img),
                        security_context=client.V1PodSecurityContext(
                            run_as_user=self.args.uid)
                    )
                ],
                restart_policy="Never",
                node_name=node
            )
            return spec

    def _podname_from_image(self, img):
        with start_action(action_type="_podname_from_image"):
            iname = '-'.join(img.split('/')[-2:])
            iname = iname.replace(':', '-')
            iname = iname.replace('_', '-')
            return iname

    def clean_completed_pods(self):
        '''Get a pod list and delete any that are in the speclist and have
        already run to completion.
        '''
        with start_action(action_type="clean_completed_pods"):
            self._destroy_pods(selective=True)

    def start_single_pod(self, spec):
        '''Run a pod, with a single container, on a particular node.
        (Assuming that the pod is itself tied to a node in the pod spec.)
        This has the effect of pulling the image for that pod onto that
        node.  The run itself is unimportant.  It returns the name of the
        created pod.
        '''
        with start_action(action_type="start_single_pod"):
            v1 = self.client
            name = self._derive_pod_name(spec)
            pod = client.V1Pod(spec=spec,
                               metadata=client.V1ObjectMeta(
                                   name=name)
                               )
            name = spec.containers[0].name
            self.logger.debug("Running pod %s" % name)
            made_pod = v1.create_namespaced_pod(self.namespace, pod)
            podname = made_pod.metadata.name
            return podname

    def _derive_pod_name(self, spec):
        '''Pod name is based on image and node.
        '''
        with start_action(action_type="_derive_pod_name"):
            return ("pp-" +
                    self._podname_from_image(spec.containers[0].image) +
                    "-" + spec.node_name.split('-')[-1])

    def run_pods(self):
        '''Run pods for all nodes.  Parallelize across nodes.
        '''
        with start_action(action_type="run_pods"):
            tlist = []
            for node in self.pod_specs:
                speclist = list(self.pod_specs[node])
                thd = Thread(target=self.run_pods_for_node,
                             args=(node, speclist))
                tlist.append(thd)
                thd.start()
            # Wait for all threads to return
            for thd in tlist:
                self.logger.debug("Waiting for thread '%s'" % thd.name)
                thd.join()

    def run_pods_for_node(self, node, speclist):
        '''Execute pods one at a time, so we don't overwhelm I/O.
        Execute this method in parallel across all nodes for best
        results.  Each node should have its own I/O, so they should all be
        busy at once.
        '''
        with start_action(action_type="run_pods_for_node"):
            self.logger.debug("Running pods for node %s" % node)
            for spec in speclist:
                name = spec.containers[0].name
                self.logger.debug("Running pod '%s' for node '%s'" % (name,
                                                                      node))
                podname = self.start_single_pod(spec)
                self.wait_for_pod(podname)

    def wait_for_pod(self, podname, delay=1, max_tries=3600):
        '''Wait for a particular pod to go into phase "Succeeded" or
        "Failed", and then delete the pod.
        Raise an exception if the delay timer expires.
        '''
        with start_action(action_type="wait_for_pod"):
            v1 = self.client
            namespace = self.namespace
            tries = 1
            while True:
                pod = v1.read_namespaced_pod(podname, namespace)
                phase = pod.status.phase
                if phase in ["Failed", "Succeeded"]:
                    if phase == "Failed":
                        self.logger.error("Pod '%s' failed" % podname)
                    self.delete_pod(podname)
                    return
                if tries >= max_tries:
                    errstr = ("Pod '%s' did not complete after " % podname +
                              "%d %d s iterations." % (max_tries, delay))
                    self.logger.error(errstr)
                    raise RuntimeError(errstr)
                pstr = "Wait %d s [%d/%d]for pod '%s' [%s]" % (
                    delay, tries, max_tries, podname, phase)
                self.logger.debug(pstr)
                time.sleep(delay)
                tries = tries + 1

    def delete_pod(self, podname):
        '''Delete a named pod.
        '''
        with start_action(action_type="delete_pod"):
            v1 = self.client
            self.logger.debug("Deleting pod %s" % podname)
            v1.delete_namespaced_pod(podname, self.namespace)



'''Class to provide namespace manipulation.
'''

import json
import time
import yaml
from eliot import start_action
from kubernetes.client.rest import ApiException
from kubernetes import client
from .. import LoggableChild
from ..utils import (make_passwd_line, make_group_lines, add_user_to_groups,
                     assemble_gids)


class LSSTNamespaceManager(LoggableChild):
    '''Class to provide namespace manipulation.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.namespace = None
        self.service_account = None  # Account for pod to run as

    def set_namespace(self, namespace):
        with start_action(action_type="set_namespace"):
            self.namespace = namespace

    def ensure_namespace(self):
        '''Here we make sure that the namespace exists, creating it if
        it does not.  That requires a ClusterRole that can list and create
        namespaces.

        If we create the namespace, we also create (if needed) a ServiceAccount
        within it to allow the user pod to spawn dask and workflow pods.

        '''
        with start_action(action_type="ensure_namespace"):
            namespace = self.namespace
            if not namespace or namespace == "default":
                raise ValueError("Will not use default namespace!")
            api = self.parent.api
            cfg = self.parent.config
            acd = 'argocd.argoproj.io/'
            ns = client.V1Namespace(
                metadata=client.V1ObjectMeta(
                    name=namespace,
                    labels={acd + 'instance': 'nublado-users'},
                    annotations={
                        acd + 'compare-options': 'IgnoreExtraneous',
                        acd + 'sync-options': 'Prune=false',
                    }))
            try:
                self.log.info(
                    "Attempting to create namespace '%s'" % namespace)
                api.create_namespace(ns)
            except ApiException as e:
                if e.status != 409:
                    estr = "Create namespace '%s' failed: %s" % (ns, str(e))
                    self.log.exception(estr)
                    raise
                else:
                    self.log.info("Namespace '%s' already exists." % namespace)
            # Wait for the namespace to actually appear before creating objects
            #  in it.
            self.wait_for_namespace()
            self.log.debug("Ensuring namespaced config maps.")
            self.ensure_namespaced_config_maps()
            if cfg.allow_dask_spawn:
                self.log.debug("Ensuring namespaced service account.")
                self.ensure_namespaced_service_account()
            if self.parent.spawner.enable_namespace_quotas:
                # By the time we need this, quota will have been set, because
                #  we needed it for options form generation.
                self.log.debug("Determining resource quota.")
                qm = self.parent.quota_mgr
                qm.ensure_namespaced_resource_quota(qm.quota)
            self.log.debug("Namespace resources ensured.")

    def def_namespaced_account_objects(self):
        '''Define K8s objects for things we need in the namespace.
        '''
        with start_action(action_type="define_namespaced_account_objects"):
            namespace = self.namespace
            username = self.parent.user.escaped_name
            account = "{}-svcacct".format(username)
            self.service_account = account
            acd = 'argocd.argoproj.io/'
            md = client.V1ObjectMeta(
                name=account,
                labels={acd + 'instance': 'nublado-users'},
                annotations={
                    acd + 'compare-options': 'IgnoreExtraneous',
                    acd + 'sync-options': 'Prune=false',
                }
            )
            svcacct = client.V1ServiceAccount(metadata=md)
            # These rules let us manipulate Dask pods, Argo Workflows, and
            #  Multus CNI interfaces
            rules = [
                client.V1PolicyRule(
                    api_groups=["argoproj.io"],
                    resources=["workflows", "workflows/finalizers"],
                    verbs=["get", "list", "watch", "update", "patch", "create",
                           "delete"]
                ),
                client.V1PolicyRule(
                    api_groups=["argoproj.io"],
                    resources=["workflowtemplates",
                               "workflowtemplates/finalizers"],
                    verbs=["get", "list", "watch"],
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["secrets"],
                    verbs=["get"]
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["pods", "pods/exec", "services", "configmaps"],
                    verbs=["get", "list", "watch", "create", "delete",
                           "update", "patch"]
                ),
                client.V1PolicyRule(
                    api_groups=[""],
                    resources=["pods/log", "serviceaccounts"],
                    verbs=["get", "list", "watch"]
                ),
            ]
            role = client.V1Role(
                rules=rules,
                metadata=md)
            rbstr = 'rbac.authorization.k8s.io'
            rolebinding = client.V1RoleBinding(
                metadata=md,
                role_ref=client.V1RoleRef(api_group=rbstr,
                                          kind="Role",
                                          name=account),
                subjects=[client.V1Subject(
                    kind="ServiceAccount",
                    name=account,
                    namespace=namespace)]
            )
            return svcacct, role, rolebinding

    def ensure_namespaced_service_account(self):
        '''Create a service account with role and rolebinding to allow it
        to manipulate resources in the namespace.
        '''
        with start_action(action_type="ensure_namespaced_service_account"):
            self.log.info("Ensuring namespaced service account.")
            namespace = self.namespace
            api = self.parent.api
            rbac_api = self.parent.rbac_api
            svcacct, role, rolebinding = self.def_namespaced_account_objects()
            account = self.service_account
            try:
                self.log.info("Attempting to create service account.")
                api.create_namespaced_service_account(
                    namespace=namespace,
                    body=svcacct)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception(("Create service account '{}' " +
                                        "in namespace '{}' " +
                                        "failed: '{}").format(account,
                                                              namespace, e))
                    raise
                else:
                    self.log.info(("Service account '{}' " +
                                   "in namespace '{}' " +
                                   "already exists.").format(account,
                                                             namespace))
            try:
                self.log.info("Attempting to create role in namespace.")
                rbac_api.create_namespaced_role(
                    namespace,
                    role)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception(("Create role '{}' " +
                                        "in namespace '{}' " +
                                        "failed: {}").format(account,
                                                             namespace, e))
                    raise
                else:
                    self.log.info(("Role '{}' already exists in " +
                                   "namespace '{}'.").format(account,
                                                             namespace))
            try:
                self.log.info("Attempting to create rolebinding in namespace.")
                rbac_api.create_namespaced_role_binding(
                    namespace,
                    rolebinding)
            except ApiException as e:
                if e.status != 409:
                    self.log.exception(("Create rolebinding '{}' " +
                                        "in namespace '{}' " +
                                        "failed: {}").format(account,
                                                             namespace, e))
                    raise
                else:
                    self.log.info("Rolebinding '%s' " % account +
                                  "already exists in '%s'." % namespace)

    def def_lab_config_maps(self):
        '''Create ConfigMaps for data we will need in the
        spawned environment.  Returns a dict with string keys which are
        the basenames of the ConfigMap-represented files, and values
        which are the in-memory K8s ConfigMap objects.
        '''
        with start_action(action_type="def_lab_config_maps"):
            cfg = self.parent.config
            auth = self.parent.authenticator
            ast = auth.cached_auth_state
            if not ast:
                errstr = "Authenticator has no cached_auth_state!"
                self.log.error(errstr)
                raise RuntimeError(errstr)
            uname = ast['claims']['uid']
            token = ast['access_token']
            claims = ast['claims']
            vol_file = cfg.volume_definition_file
            with open(vol_file, "r") as fp:
                vdata = fp.read()
            v_nm = 'mountpoints'
            vol_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=v_nm),
                data={v_nm: vdata})
            self.log.debug("Created volume configmap '{}'".format(v_nm))
            pw_file = cfg.base_passwd_file
            with open(pw_file, "r") as fp:
                b_pdata = fp.read().rstrip() + '\n'
            p_nm = 'passwd'
            pline = make_passwd_line(claims)
            pdata = b_pdata + pline
            pw_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=p_nm),
                data={p_nm: pdata})
            grp_file = cfg.base_group_file
            with open(grp_file, "r") as fp:
                b_gdata = fp.read().rstrip() + '\n'
            gdata = add_user_to_groups(uname, b_gdata)
            g_nm = 'group'
            glines = make_group_lines(
                claims, strict_ldap=cfg.strict_ldap_groups)
            for gl in glines:
                gdata = gdata + gl
            grp_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=g_nm),
                data={g_nm: gdata})
            a_nm = 'access-token'
            # Use fqdn to uniquely identify instance.
            mt_nm = cfg.fqdn + "-token"
            at_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=a_nm),
                data={mt_nm: token})
            s_nm = 'shadow'
            s_pw_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=s_nm),
                data={s_nm: self._pwconv(pdata)})
            sg_nm = 'gshadow'
            s_grp_configmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name=sg_nm),
                data={sg_nm: self._grpconv(gdata)})
            cm_map = {
                v_nm: vol_configmap,
                p_nm: pw_configmap,
                s_nm: s_pw_configmap,
                g_nm: grp_configmap,
                sg_nm: s_grp_configmap,
                a_nm: at_configmap
            }
            return cm_map

    def ensure_dask_configmap(self, daskconfig):
        '''This one can't be called until spawn time, once we know the
        parameters (image, size, etc.).  Daskconfig is a dict with string
        keys.
        '''
        with start_action(action_type='ensure_dask_configmap'):
            namespace = self.namespace
            api = self.parent.api
            cfg = self.parent.config
            if not cfg.allow_dask_spawn:
                self.log.warning(
                    "Dask spawns disallowed.  Not making configmap.")
                return
            dask_config = self._def_dask_configmap(daskconfig)
            try:
                self.log.info(
                    (("Attempting to create configmap 'dask-config' " +
                      "in {}").format(namespace)))
                api.create_namespaced_config_map(namespace, dask_config)
            except ApiException as e:
                if e.status != 409:
                    estr = "Create configmap failed: {}".format(e)
                    self.log.exception(estr)
                    raise
                else:
                    self.log.info("Configmap already exists.")

    def _def_dask_configmap(self, daskconfig):
        with start_action(action_type='_def_dask_configmap'):
            d_yaml = self._def_dask_yaml(daskconfig)
            # Same per-instance strategy as access_token
            fqdn = self.parent.config.fqdn
            fn = fqdn + "-dask_worker.example.yml"
            cmap = client.V1ConfigMap(
                metadata=client.V1ObjectMeta(name='dask-config'),
                data={fn: d_yaml})
            return cmap

    def _def_dask_yaml(self, daskconfig):
        with start_action(action_type='_def_dask_yaml'):
            cfg = self.parent.config
            vm = self.parent.volume_mgr
            ast = daskconfig['auth_state']
            claims = ast['claims']
            uname = claims['uid']
            uid = int(ast['uid'])
            uid_str = str(ast['uid'])
            debug = ''
            instance_name = ''
            if daskconfig.get('instance_name'):
                instance_name = daskconfig['instance_name']
            if daskconfig['debug']:
                debug = 'true'
            e_grps = assemble_gids(
                claims, strict_ldap=cfg.strict_ldap_groups)
            resources = daskconfig['resources']
            image = daskconfig['image']
            cmd = '/opt/lsst/software/jupyterlab/provisionator.bash'
            mounts = vm.get_volume_mount_list()
            vols = vm.get_volume_list()
            run_as = 769
            if cfg.lab_no_sudo:
                run_as = uid
            dask_ctr = {'image': image,
                        'imagePullPolicy': 'Always',
                        'args': [cmd],
                        'name': 'dask',
                        'env': [
                            {'name': 'DASK_WORKER',
                             'value': 'true'},
                            {'name': 'DEBUG',
                             'value': debug},
                            {'name': 'EXTERNAL_GROUPS',
                             'value': e_grps},
                            {'name': 'EXTERNAL_UID',
                             'value': uid_str},
                            {'name': 'JUPYTERHUB_USER',
                             'value': uname},
                            {'name': 'INSTANCE_NAME',
                             'value': instance_name},
                            {'name': 'CPU_LIMIT',
                             'value': resources['limits']['cpu']},
                            {'name': 'MEM_LIMIT',
                             'value': resources['limits']['memory']},
                            {'name': 'CPU_GUARANTEE',
                             'value': resources['requests']['cpu']},
                            {'name': 'MEM_GUARANTEE',
                             'value': resources['requests']['memory']}
                        ],  # env
                        'resources': resources,
                        'volumeMounts': mounts}
            if cfg.restrict_dask_nodes:
                dask_ctr['nodeSelector'] = {'dask': 'ok'}
            self.log.error("Dask container: {}".format(dask_ctr))
            self.log.error("Volumes: {}".format(vols))
            dask_spec = {
                'restartPolicy': 'Never',
                'securityContext': {
                    'runAsUser': run_as,
                    'runAsGroup': run_as
                }
            }
            self.log.error("Dask_spec_pre: {}".format(dask_spec))
            dask_spec['containers'] = [dask_ctr]
            self.log.error("Dask_spec_pre2: {}".format(dask_spec))
            dask_spec['volumes'] = vols
            self.log.error("Dask_spec_post: {}".format(dask_spec))
            dask_dict = {'kind': 'Pod',
                         'metadata': {
                             'labels': {
                                 'dask': 'ok'
                             }
                         },
                         'spec': dask_spec}
            self.log.error('Dask pod object: {}'.format(dask_dict))
            c_yaml = yaml.dump(dask_dict)
            self.log.error('Dask YAML: {}'.format(c_yaml))
            return c_yaml

    def _pwconv(self, pwdata):
        pwlines = pwdata.split('\n')
        spwlines = []
        for pl in pwlines:
            pll = pl.strip()
            if not pll:
                continue
            unm = pll.split(':')[0]
            # 18000 is late March 2019; doesn't really matter
            spl = "{}:*:18000:0:99999:7:::\n".format(unm)
            spwlines.append(spl)
        spwdata = ''.join(spwlines)  # They already have newlines
        return spwdata

    def _grpconv(self, grpdata):
        grplines = grpdata.split('\n')
        sglines = []
        for grp in grplines:
            grpl = grp.strip()
            if not grpl:
                continue
            flds = grpl.split(':')
            gnm = flds[0]
            gmem = flds[-1]
            sgl = "{}:!::{}\n".format(gnm, gmem)
            sglines.append(sgl)
        sgdata = ''.join(sglines)  # They already have newlines
        return sgdata

    def ensure_namespaced_config_maps(self):
        with start_action(action_type="ensure_namespaced_config_maps"):
            namespace = self.namespace
            api = self.parent.api
            cm_map = self.def_lab_config_maps()
            for cm in cm_map:
                try:
                    self.log.info(
                        (("Attempting to create configmap {} " +
                          "in {}").format(cm, namespace)))
                    api.create_namespaced_config_map(namespace, cm_map[cm])
                except ApiException as e:
                    if e.status != 409:
                        estr = "Create configmap failed: {}".format(e)
                        self.log.exception(estr)
                        raise
                    else:
                        self.log.info("Configmap already exists.")

    def wait_for_namespace(self, timeout=30):
        '''Wait for namespace to be created.'''
        with start_action(action_type="wait_for_namespace"):
            namespace = self.namespace
            for dl in range(timeout):
                self.log.debug("Checking for namespace " +
                               "{} [{}/{}]".format(namespace, dl, timeout))
                nl = self.parent.api.list_namespace(timeout_seconds=1)
                for ns in nl.items:
                    nsname = ns.metadata.name
                    if nsname == namespace:
                        self.log.debug("Namespace {} found.".format(namespace))
                        return
                    self.log.debug(
                        "Namespace {} not present yet.".format(namespace))
                time.sleep(1)
            raise RuntimeError(
                "Namespace '{}' not created in {} seconds!".format(namespace,
                                                                   timeout))

    def maybe_delete_namespace(self):
        '''Here we try to delete the namespace.  If it has no non-dask
        running pods, and it's not the default namespace, we can delete it."

        This requires a cluster role that can delete namespaces.'''
        with start_action(action_type="maybe_delete_namespace"):
            self.log.debug("Attempting to delete namespace.")
            namespace = self.namespace
            if not namespace or namespace == "default":
                raise RuntimeError("Cannot delete default namespace!")
            podlist = self.parent.api.list_namespaced_pod(namespace)
            clear_to_delete = True
            if podlist and podlist.items:
                clear_to_delete = self._check_pods(podlist.items)
            if not clear_to_delete:
                self.log.info("Not deleting namespace '%s'" % namespace)
                return False
            self.log.info("Deleting namespace '%s'" % namespace)
            self.parent.api.delete_namespace(namespace)
            return True

    def _check_pods(self, items):
        with start_action(action_type="_check_pods"):
            namespace = self.namespace
            for i in items:
                if i and i.status:
                    phase = i.status.phase
                    if (phase == "Running" or phase == "Unknown"
                            or phase == "Pending"):
                        pname = i.metadata.name
                        if pname.startswith("dask-"):
                            self.log.debug(
                                ("Abandoned dask pod '{}' can be " +
                                 "reaped.").format(pname))
                            # We can murder abandoned dask pods
                            continue
                        self.log.warning(("Pod in state '{}'; cannot delete " +
                                          "namespace '{}'.").format(phase,
                                                                    namespace))
                        return False
            # FIXME check on workflows as well.
            return True

    def destroy_namespaced_resource_quota(self):
        '''Remove quotas from namespace.
        You don't usually have to call this, since it will get
        cleaned up as part of namespace deletion.
        '''
        with start_action(action_type="destroy_namespaced_resource_quota"):
            namespace = self.get_user_namespace()
            api = self.parent.api
            qname = "quota-" + namespace
            dopts = client.V1DeleteOptions()
            self.log.info("Deleting resourcequota '%s'" % qname)
            api.delete_namespaced_resource_quota(qname, namespace, dopts)

    def delete_namespaced_svcacct_objs(self):
        '''Remove service accounts, roles, and rolebindings from namespace.
        You don't usually have to call this, since they will get
         cleaned up as part of namespace deletion.
        '''
        with start_action(action_type="delete_namespaced_svcacct_objs"):
            namespace = self.namespace
            account = self.service_account
            if not account:
                self.log.info("Service account not defined.")
                return
            dopts = client.V1DeleteOptions()
            self.log.info("Deleting service accounts/role/rolebinding " +
                          "for %s" % namespace)
            self.parent.rbac_api.delete_namespaced_role_binding(
                account,
                namespace,
                dopts)
            self.parent.rbac_api.delete_namespaced_role(
                account,
                namespace,
                dopts)
            self.parent.api.delete_namespaced_service_account(
                account,
                namespace,
                dopts)

    def dump(self):
        '''Return dict for pretty-printing.
        '''
        nd = {"namespace": self.namespace,
              "service_account": self.service_account,
              "parent": str(self.parent)}
        return nd

    def toJSON(self):
        return json.dumps(self.dump())



'''LSST JupyterHub utilities and helpers.

These implement the LSST-specific tooling for the LSST Science
Platform Notebook Aspect.  The repo scanner looks for Docker images in
a repository with a particular tag format; the prepuller pulls a
subset of those images to each node.  The reaper removes images past a
certain age, based on the tag format.  The LSST Manager class provides
a hierarchy of objects that hold LSST-specific configuration and logic
for spawning JupyterLab pods, and the spawner and authenticators
provide the pod spawner and the LSST-supported authentication methods
and logic.  Convenience functions are in 'utils' and JupyterHub
configuration convenience functions are in 'config_helpers'.
'''
from .singleton import Singleton
from .loggable import Loggable, LoggableChild
from .scanrepo import ScanRepo, SingletonScanner, Prepuller, Reaper
from .lsstmgr import LSSTMiddleManager
from .spawner import LSSTSpawner
from .authenticator.lsstjwtauth import LSSTJWTAuthenticator
from .utils import (rreplace, sanitize_dict, get_execution_namespace,
                    make_logger, str_bool, str_true, listify, intify,
                    floatify, list_duplicates, list_digest,
                    get_access_token, parse_access_token, assemble_gids,
                    make_passwd_line, make_group_lines, add_user_to_groups,
                    get_supplemental_gids, resolve_groups)
from .config import LSSTConfig
from ._version import __version__

__all__ = [LSSTMiddleManager, Prepuller, Reaper, ScanRepo, Singleton,
           SingletonScanner, LSSTSpawner, LSSTJWTAuthenticator,
           rreplace, sanitize_dict, get_execution_namespace,
           make_logger, str_bool, str_true, listify, intify, floatify,
           list_duplicates, list_digest, get_access_token,
           parse_access_token, assemble_gids, make_passwd_line,
           make_group_lines, add_user_to_groups, get_supplemental_gids,
           resolve_groups, LSSTConfig, Loggable, LoggableChild, __version__]



'''The LSSTMiddleManager is the only exported class: all the managers that
actually manage things report to it.  However, we expose some of the functions
from the group manager.
'''
from .middlemanager import LSSTMiddleManager

__all__ = [LSSTMiddleManager]



from .singletonscanner import SingletonScanner


def prime_repo_cache(cfg):
    scr = SingletonScanner(debug=cfg.debug,
                           json=True,
                           host=cfg.lab_repo_host,
                           owner=cfg.lab_repo_owner,
                           name=cfg.lab_repo_name,
                           cachefile=cfg.prepuller_cachefile,
                           experimentals=cfg.prepuller_experimentals,
                           dailies=cfg.prepuller_dailies,
                           weeklies=cfg.prepuller_weeklies,
                           releases=cfg.prepuller_releases)
    scr.scan()



'''LSST JupyterHub utility functions.
'''
from .utils import (rreplace, sanitize_dict, get_execution_namespace,
                    make_logger, str_bool, str_true, listify, intify,
                    floatify, list_duplicates, list_digest,
                    get_access_token, parse_access_token, assemble_gids,
                    get_fake_gid, make_passwd_line, make_group_lines,
                    add_user_to_groups, get_supplemental_gids, resolve_groups)
all = [rreplace, sanitize_dict, get_execution_namespace, make_logger,
       str_bool, str_true, listify, intify, floatify, list_duplicates,
       list_digest, get_access_token, parse_access_token, assemble_gids,
       get_fake_gid, make_passwd_line, make_group_lines, add_user_to_groups,
       get_supplemental_gids, resolve_groups]



'''LSST Authenticator to use JWT token present in request headers.
'''
import asyncio
from eliot import start_action
from jwtauthenticator.jwtauthenticator import JSONWebTokenAuthenticator
from .lsstauth import LSSTAuthenticator
from .lsstjwtloginhandler import LSSTJWTLoginHandler
from ..utils import make_logger, resolve_groups


class LSSTJWTAuthenticator(LSSTAuthenticator, JSONWebTokenAuthenticator):

    def __init__(self, *args, **kwargs):
        self.log = make_logger()
        self.log.debug("Creating LSSTJWTAuthenticator")
        # Superclass gives us the LSST Manager
        super().__init__(*args, **kwargs)
        self.auth_refresh_age = 900
        self.header_name = "X-Portal-Authorization"
        self.header_is_authorization = True
        self.username_claim_field = 'uid'
        self.cached_auth_state = {}

    def get_handlers(self, app):
        '''Install custom handlers.
        '''
        with start_action(action_type="get_handlers"):
            return [
                (r'/login', LSSTJWTLoginHandler),
            ]

    def logout_url(self, base_url):
        '''Returns the logout URL for JWT.
        '''
        with start_action(action_type="logout_url"):
            return self.lsst_mgr.config.jwt_logout_url

    async def refresh_user(self, user, handler):
        '''Delegate to login handler, if this happens in the login
        '''
        # We don't want to do this anywhere but on the login handler.
        #  It's cheating, but we'll just check to see if there is
        #  a custom method for refresh_user on the handler and call it
        #  if so.  That's true for the LSST JWT Authenticator case.
        with start_action(action_type="refresh_user_lsstjwtauth"):
            uname = user.escaped_name
            self.log.debug(
                "Entering lsstjwtauth refresh_user() for '{}'".format(uname))
            self.log.debug(
                "Calling superclass refresh_user for '{}'.".format(uname))
            _ = await super().refresh_user(user, handler)
            self.log.debug(
                "Returned from  superclass refresh_user for '{}'.".format(
                    uname))
            if hasattr(handler, 'refresh_user'):
                self.log.debug("Handler has refresh_user too.")
                self.log.debug(
                    "Calling handler's refresh_user() for '{}'.".format(uname))
                _ = await handler.refresh_user(user, handler)
                self.log.debug(
                    "Returned from handler refresh_user for '{}'.".format(
                        uname))
            # Set uid and group_map
            # Add 'uid' and 'group_map' to auth_state per lsstauth.py
            strict_ldap = self.lsst_mgr.config.strict_ldap_groups
            ast = await user.get_auth_state()
            claims = ast['claims']
            ast['uid'] = claims['uidNumber']
            ast['group_map'] = resolve_groups(claims, strict_ldap)
            await user.save_auth_state(ast)
            self.cached_auth_state = ast  # For ease of synchronous access
            return ast



import argparse
import json
from ..config import LSSTConfig
from eliot import start_action


def parse_args(cfg=LSSTConfig(),
               desc="Get list of Lab Images for display or prepulling",
               component="scanner"):
    '''Parse command-line arguments.
    '''
    with start_action(action_type="parse_args"):
        parser = argparse.ArgumentParser(description=desc)
        dbg = cfg.debug
        lrh = cfg.lab_repo_host
        lro = cfg.lab_repo_owner
        lrn = cfg.lab_repo_name
        kex = cfg.prepuller_experimentals
        kdl = cfg.prepuller_dailies
        kwk = cfg.prepuller_weeklies
        krl = cfg.prepuller_releases
        ppc = cfg.prepuller_cachefile
        pto = cfg.prepuller_timeout
        ppn = cfg.prepuller_namespace
        lid = cfg.lab_uid
        cmd = cfg.prepuller_command
        component = component.lower()
        allowed_components = ["scanner", "reaper", "prepuller"]
        if component not in allowed_components:
            raise ValueError(
                "Component {} not in {}!".format(component,
                                                 allowed_components))
        # These are larger for the reaper; there should be a number of not-
        #  displayed images we haven't destroyed yet.
        if component == "reaper":
            kex = cfg.reaper_keep_experimentals
            kdl = cfg.reaper_keep_dailies
            kwk = cfg.reaper_keep_weeklies

        parser.add_argument("-d", "--debug", action="store_true",
                            help="enable debugging [{}]".format(dbg),
                            default=dbg)
        parser.add_argument("-f", "--cachefile",
                            help="Cachefile for results [{}]".format(ppc),
                            default=ppc)
        parser.add_argument("-r", "--repo", "--repository",
                            help="Docker repository host [{}]".format(lrh),
                            default=lrh)
        parser.add_argument("-o", "--owner", "--organization", "--org",
                            help="repository owner [{}]".format(lro),
                            default=lro)
        parser.add_argument("-n", "--name",
                            help="repository name [{}]".format(lrn),
                            default=lrn)
        parser.add_argument("-q", "--dailies", "--daily", "--quotidian",
                            type=int,
                            help="# of daily builds to keep [{}]".format(kdl),
                            default=kdl)
        parser.add_argument("-w", "--weeklies", "--weekly", type=int,
                            help="# of weekly builds [{}]".format(kwk),
                            default=kwk)
        parser.add_argument("-e", "--experimentals", "--experimental",
                            "--exp", type=int,
                            help=(
                                "# of experimental builds to keep [{}]".format(
                                    kex)),
                            default=kex)
        if component != "reaper":
            parser.add_argument("-b", "--releases", "--release", type=int,
                                help=(
                                    "# of release builds to keep [{}]".format(
                                        krl)),
                                default=krl)
        parser.add_argument("-p", "--port", help="Repository port [443 for" +
                            " secure, 80 for insecure]",
                            default=None)
        parser.add_argument("-i", "--insecure", "--no-tls", "--no-ssl",
                            help="Do not use TLS to connect [False]",
                            action='store_true',
                            default=False)
        if component != "reaper":
            parser.add_argument("-l", "--list", "--list-images",
                                "--image-list",
                                help=("Use supplied comma-separated list in" +
                                      " addition to repo scan"),
                                default=None)
            parser.add_argument("-s", "--sort", "--sort-field", "--sort-by",
                                help="Field to sort results by [name]",
                                default="name")
            parser.add_argument("--recommended",
                                help=(
                                    "Pull image w/tag 'recommended' [True]"),
                                type=bool,
                                default=True)
        if component == "scanner":
            parser.add_argument("-j", "--json", help="Emit results as JSON, " +
                                " not sourceable shell fragment [False]",
                                action='store_true',
                                default=False)
        if component == "prepuller":
            parser.add_argument("--no-scan", action="store_true",
                                help=("Do not do repo scan (only useful in" +
                                      " conjunction with --list)"),
                                default=False)
            parser.add_argument("--command",
                                help=("JSON representation of command " +
                                      "to run when image is run as " +
                                      "prepuller " +
                                      "[{}]".format(str(cmd))),
                                default=cmd)
            parser.add_argument("-t", "--timeout",
                                help=(
                                    "Seconds allowed for process to " +
                                    "complete " +
                                    "(-1 for no timeout) [{}]".format(pto)),
                                type=int,
                                default=pto)
            parser.add_argument("-u", "--uid",
                                help=("UID to run as [{}]".format(lid)),
                                type=int,
                                default=lid)
            parser.add_argument("--namespace",
                                help="Kubernetes namespace [{}]".format(ppn),
                                default=ppn)
        if component == "reaper":
            parser.add_argument("--dry-run", action='store_true',
                                help="Don't actually delete images",
                                default=False)
        results = parser.parse_args()
        results.path = ("/v2/repositories/" + results.owner + "/" +
                        results.name + "/tags/")
        if (hasattr(results, "list") and results.list):
            results.list = list(set(results.list.split(',')))
        if (hasattr(results, "command") and results.command and
                type(results.command) is str):
            results.command = json.loads(results.command)
        return results



import datetime
import threading
import time
from eliot import start_action
from .scanrepo import ScanRepo
from ..singleton import Singleton


class SingletonScanner(ScanRepo, metaclass=Singleton):
    '''Singleton Object to hold rate-limited scanner.
    '''

    def __init__(self, **kwargs):
        min_refresh_time = kwargs.get('min_refresh_time', 60)
        max_cache_age = kwargs.get('max_cache_age', 600)
        if max_cache_age is None:
            max_cache_age = 600
        # Now remove them from kwargs before superclass init
        for karg in ['min_refresh_time', 'max_cache_age']:
            if karg in kwargs:
                kwargs.pop(karg)
        super().__init__(**kwargs)
        self.min_refresh_time = min_refresh_time
        self.max_cache_age = max_cache_age
        self.last_updated = datetime.datetime(1970, 1, 1)  # The Epoch
        self.scanning = False
        self.lock = threading.RLock()
        if max_cache_age < min_refresh_time:
            max_cache_age = 2 * min_refresh_time
            self.max_cache_age = max_cache_age
            self.logger.error("Nonsensical cache age/refresh time ratio.")
            self.logger.warning("Setting max_age to %ds." % max_cache_age)
        thd = threading.Thread(target=self.scan)
        self.logger.info("Starting background scan.")
        thd.start()

    def scan(self):
        '''Execute repo scan.
        '''
        with start_action(action_type="scan"):
            _timeout = 300
            _initialdelay = 0.2
            _maxdelay = 15
            _sofar = 0
            delay = _initialdelay
            if self.scanning:
                while not self._results:
                    # If there have been no results, wait up to
                    # _timeout seconds for them.
                    self.logger.debug(
                        "Scan in progress; waiting {}s for results.".format(
                            delay))
                    time.sleep(delay)
                    _sofar += delay
                    if _sofar >= _timeout:
                        raise ValueError("Scan in progress never finished.")
                    delay *= 2
                    if delay > _maxdelay:
                        delay = _maxdelay
                    if _sofar + delay > _timeout:
                        delay = _timeout - _sofar
                    return  # Use cached (possibly stale) version if
                    # results exist.
            self.scanning = True
            now = datetime.datetime.utcnow()
            mt = self.min_refresh_time
            min_delay = datetime.timedelta(seconds=mt)
            if (now - self.last_updated < min_delay):
                self.logger.warning("%ds not elapsed; not rescanning." % mt)
                self.scanning = False
                return
            self.logger.info("Rescanning.")
            with self.lock:
                super().scan()
                self.last_updated = now
                self.scanning = False

    def _scan_if_needed(self):
        with start_action(action_type="_scan_if_needed"):
            now = datetime.datetime.utcnow()
            max_age = datetime.timedelta(seconds=self.max_cache_age)
            last_updated = self.last_updated
            if ((now - last_updated) > max_age):
                self.logger.info("Scan data has expired.")
                if self.scanning:
                    self.logger.info("Waiting for scan results.")
                    count = 0
                    while True:
                        time.sleep(0.2)
                        count = count + 1
                        if not (count % 50):
                            self.logger.debug(
                                "Waiting for data: {}s".format(count / 5))
                        if self.data:
                            break
                else:
                    self.logger.info("Beginning new scan.")
                    self.scan()
                self.logger.debug("Exiting _scan_if_needed() wait loop.")

    def get_data(self):
        '''Return repo data.
        '''
        with start_action(action_type="get_data"):
            self._scan_if_needed()
            return self.data

    def get_all_tags(self):
        '''Return all tags in repo.
        '''
        with start_action(action_type="get_all_tags"):
            self._scan_if_needed()
            return self._all_tags

    def get_all_scan_results(self):
        '''Return results from repository scan as dict.
        '''
        with start_action(action_type="get_all_scan_results"):
            self._scan_if_needed()
            return self._results_map

    def extract_image_info(self):
        '''Get info for all images.
        '''
        with start_action(action_type="extract_image_info"):
            self._scan_if_needed()
            return super().extract_image_info()

    def report(self):
        '''Report results of scan.
        '''
        with start_action(action_type="report"):
            self._scan_if_needed()
            return super().report()



import json
from .. import LoggableChild
from .. import Singleton
from kubernetes.client import CoreV1Api, RbacAuthorizationV1Api
from kubernetes.config import load_incluster_config, load_kube_config
from kubernetes.config.config_exception import ConfigException


class LSSTAPIManager(LoggableChild, metaclass=Singleton):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            load_incluster_config()
        except ConfigException:
            self.log.warning("In-cluster config failed! Falling back.")
            try:
                load_kube_config()
            except ValueError as exc:
                self.log.error("Still errored: {}".format(exc))
        rbac_api = kwargs.pop('rbac_api', None)
        if not rbac_api:
            rbac_api = RbacAuthorizationV1Api()
        self.rbac_api = rbac_api
        api = kwargs.pop('api', None)
        if not api:
            api = CoreV1Api()
        self.api = api

    def dump(self):
        '''Return contents dict for aggregation and pretty-printing.
        '''
        ad = {"parent": str(self.parent),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              }
        return ad

    def toJSON(self):
        return json.dumps(self.dump())



import json
import logging
import os
import sys
from eliot import to_file
from eliot.stdlib import EliotHandler
from jupyter_client.localinterfaces import public_ips
from .. import Singleton
from ..utils import (str_bool, listify, intify, floatify, make_logger,
                     get_execution_namespace, sanitize_dict)


class LSSTConfig(metaclass=Singleton):
    '''LSSTConfig is a Singleton bag of attributes to hold
    important-to-the-LSST JupyterHub environment variables and default
    values.
    '''

    def __init__(self, *args, **kwargs):
        default_level = logging.WARNING
        # we can't check this prior to the config existing...
        self.debug = str_bool(os.getenv('DEBUG', False))
        if self.debug:
            default_level = logging.DEBUG
        override_root = str_bool(os.getenv('ELIOT_ROOT_LOGGING', False))
        self._setup_logger(level=default_level, override_root=override_root)
        self.source = kwargs.pop('source', 'environment')
        if self.source != 'environment':
            raise ValueError(
                "'environment' is currently only allowed config source!")
        # To make it easy to read settings from a document later on...
        if self.source == 'environment':
            self.load_from_environment()
        self.create_derived_settings()

    def eliotify_root_logger(self, level):
        rootlogger = logging.getLogger()
        rootlogger.handlers = [EliotHandler()]
        loggers = []
        lrmld = logging.root.manager.loggerDict
        loggers = loggers + [logging.getLogger(name) for name in lrmld]
        rootlogger.debug("Removing all existing log handlers.")
        for lgr in loggers:
            rootlogger.debug(
                "Removing log handlers for '{}'".format(lgr.name))
            lgr.handlers = []
        if self.debug:
            # Reset root logger
            if level > logging.DEBUG:
                rootlogger.warn("Setting root logging level to DEBUG.")
                rootlogger.setLevel(logging.DEBUG)
        rootlogger.debug(
            "Setting default log level to '{}'".format(level))
        rootlogger.setLevel(level)
        # There are some things we just don't want to log at DEBUG
        chattycathies = {'kubernetes': logging.WARNING,
                         'urllib3': logging.WARNING,
                         'JupyterHub': logging.INFO}
        for k in chattycathies:
            v = chattycathies[k]
            lgr = logging.getLogger(k)
            el = lgr.getEffectiveLevel()
            if el < v:
                rootlogger.debug(
                    "Logger {} @level {} -> {}.".format(k, el, v))
                lgr.setLevel(v)

    def _setup_logger(self, level=logging.WARNING, override_root=False):
        # In Python 3.8 we could use force=True to force the root logger
        #  to override existing defaults
        #
        # Don't eliot-log this or load_from_environment or
        #  create_derived_settings, because we initialize an LSSTConfig when
        #  we import jupyterhubutils, and we don't necessarily want to
        #  spam everyone about it
        # Add our formats
        fstr = '[%(levelname).1s %(asctime)s.%(msecs).03d'
        fstr += ' %(module)s:%(funcName)s:%(lineno)d] %(message)s'
        dstr = '%Y-%m-%d %H:%M:%S'
        self.log_format = fstr
        self.log_datefmt = dstr
        self.log_level = level
        to_file(sys.stderr)
        if override_root:
            self.eliotify_root_logger(level=level)
        self.log = make_logger(level=level)

    def load_from_environment(self):
        '''Populate attributes from environment variables.
        '''
        # We already checked DEBUG in __init__
        # FQDN should uniquely identify an instance
        self.fqdn = os.getenv("FQDN") or 'localhost'
        # Authentication parameters
        self.jwt_logout_url = os.getenv("LOGOUT_URL") or '/logout'
        self.strict_ldap_groups = str_bool(os.getenv('STRICT_LDAP_GROUPS'))
        # Settings for Options Form
        self.form_selector_title = (os.getenv('LAB_SELECTOR_TITLE') or
                                    'Container Image Selector')
        self.form_template = (
            os.getenv('OPTIONS_FORM_TEMPLATE') or
            ('/opt/lsst/software/jupyterhub/templates/' +
             'options_form.template.html'))
        self.form_sizelist = (listify(
            os.getenv('OPTIONS_FORM_SIZELIST')) or
            ['tiny', 'small', 'medium', 'large'])
        self.max_scan_delay = intify(os.getenv('MAX_SCAN_DELAY'), 300)
        self.initial_scan_interval = floatify(
            os.getenv('INITIAL_SCAN_INTERVAL'), 0.2)
        self.max_scan_interval = floatify(
            os.getenv('MAX_SCAN_INTERVAL'), 5.0)
        self.tiny_cpu_max = floatify(os.getenv('TINY_CPU_MAX'), 0.5)
        self.mb_per_cpu = intify(os.getenv('MB_PER_CPU'), 2048)
        self.size_index = intify(os.getenv('SIZE_INDEX'), 1)
        # Settings for Quota Manager
        self.resource_map = (os.getenv('RESOURCE_MAP') or
                             '/opt/lsst/software/jupyterhub/resources/' +
                             'resourcemap.json')
        self.max_dask_workers = int(os.getenv('MAX_DASK_WORKERS', 25))
        # Settings for Volume Manager
        self.volume_definition_file = (
            os.getenv('VOLUME_DEFINITION_FILE') or
            ('/opt/lsst/software/jupyterhub/' +
             'mounts/mountpoints.json'))
        self.base_passwd_file = (
            os.getenv('BASE_PASSWD_DEFINITION_FILE') or
            '/opt/lsst/software/jupyterhub/base-files/passwd')
        self.base_group_file = (
            os.getenv('BASE_GROUP_DEFINITION_FILE') or
            '/opt/lsst/software/jupyterhub/base-files/group')
        # Hub settings for Lab and Workflow spawning
        self.lab_no_sudo = str_bool(os.getenv('LAB_NO_SUDO'))
        self.lab_uid = intify(os.getenv('LAB_UID'), 769)
        self.lab_gid = intify(os.getenv('LAB_GID'), 769)
        self.lab_fs_gid = intify(os.getenv('LAB_FS_GID'), self.lab_gid)
        self.lab_default_image = (os.getenv('LAB_IMAGE') or
                                  "lsstsqre/sciplat-lab:latest")
        self.lab_size_range = os.getenv('LAB_SIZE_RANGE') or '4.0'
        self.cull_timeout = os.getenv('LAB_CULL_TIMEOUT') or '64800'
        self.cull_policy = os.getenv('LAB_CULL_POLICY') or 'idle:remote'
        self.allow_dask_spawn = str_bool(os.getenv('ALLOW_DASK_SPAWN'))
        self.restrict_dask_nodes = os.getenv('RESTRICT_DASK_NODES')
        self.restrict_lab_nodes = os.getenv('RESTRICT_LAB_NODES')
        self.lab_nodejs_max_mem = os.getenv('LAB_NODEJS_MAX_MEM') or '6144'
        self.external_hub_url = os.getenv('EXTERNAL_HUB_URL')
        self.hub_route = os.getenv('HUB_ROUTE') or '/'
        self.external_instance_url = os.getenv('EXTERNAL_INSTANCE_URL')
        self.firefly_route = os.getenv('FIREFLY_ROUTE') or '/firefly'
        self.js9_route = os.getenv('JS9_ROUTE') or '/js9'
        self.api_route = os.getenv('API_ROUTE') or '/api'
        self.tap_route = os.getenv('TAP_ROUTE') or '/api/tap'
        self.soda_route = os.getenv('SODA_ROUTE') or '/api/image/soda'
        self.workflow_route = os.getenv('WORKFLOW_ROUTE') or '/wf'
        self.external_firefly_url = os.getenv('EXTERNAL_FIREFLY_URL')
        self.external_js9_url = os.getenv('EXTERNAL_JS9_URL')
        self.external_api_url = os.getenv('EXTERNAL_API_URL')
        self.external_tap_url = os.getenv('EXTERNAL_TAP_URL')
        self.external_soda_url = os.getenv('EXTERNAL_SODA_URL')
        self.external_workflow_url = os.getenv('EXTERNAL_WORKFLOW_URL')
        self.auto_repo_urls = os.getenv('AUTO_REPO_URLS')
        # Prepuller settings
        self.lab_repo_owner = os.getenv('LAB_REPO_OWNER') or 'lsstsqre'
        self.lab_repo_name = os.getenv('LAB_REPO_NAME') or 'sciplat-lab'
        self.lab_repo_host = os.getenv('LAB_REPO_HOST') or 'hub.docker.com'
        self.prepuller_namespace = (os.getenv('PREPULLER_NAMESPACE') or
                                    get_execution_namespace())
        self.prepuller_experimentals = intify(
            os.getenv('PREPULLER_EXPERIMENTALS'), 0)
        self.prepuller_dailies = intify(os.getenv('PREPULLER_DAILIES'), 3)
        self.prepuller_weeklies = intify(
            os.getenv('PREPULLER_WEEKLIES'), 2)
        self.prepuller_releases = intify(
            os.getenv('PREPULLER_RELEASES'), 1)
        self.prepuller_cachefile = os.getenv('PREPULLER_CACHEFILE',
                                             '/tmp/repo-cache.json')
        self.prepuller_timeout = os.getenv('PREPULLER_TIMEOUT', 3300)
        cstr = "echo \"Prepuller for $(hostname) completed at $(date).\""
        self.prepuller_command = ["/bin/sh", "-c", cstr]
        prp_cmd = os.getenv('PREPULLER_COMMAND_JSON')
        if prp_cmd:
            self.prepuller_command = json.loads(prp_cmd)
        # Reaper settings
        self.reaper_keep_experimentals = intify(
            os.getenv('REAPER_KEEP_EXPERIMENTALS'), 10)
        self.reaper_keep_dailies = intify(
            os.getenv('REAPER_KEEP_DAILIES'), 15)
        self.reaper_keep_weeklies = intify(
            os.getenv('REAPER_KEEP_WEEKLIES'), 78)
        # Fileserver settings
        self.fileserver_host = (os.getenv('EXTERNAL_FILESERVER_IP') or
                                os.getenv('FILESERVER_SERVICE_HOST'))
        # Reaper settings
        self.reaper_user = os.getenv('IMAGE_REAPER_USER')
        self.reaper_password = os.getenv('IMAGE_REAPER_PASSWORD')
        # Hub internal settings
        my_ip = public_ips()[0]
        self.instance_name = os.getenv('INSTANCE_NAME')
        self.hub_host = os.getenv('HUB_SERVICE_HOST') or my_ip
        self.hub_api_port = os.getenv('HUB_SERVICE_PORT_API') or 8081
        self.proxy_host = os.getenv('PROXY_SERVICE_HOST') or my_ip
        self.proxy_api_port = os.getenv('PROXY_SERVICE_PORT_API') or 8001
        self.session_db_url = os.getenv('SESSION_DB_URL')
        # Pod multicast (used with SAL)
        self.enable_multus = str_bool(os.getenv('ENABLE_MULTUS'))
        # DDS domain (used with SAL)
        self.lab_dds_domain = os.getenv('LSST_DDS_DOMAIN')
        # These have to be set post-initialization to avoid a circular
        #  dependency.
        self.authenticator_class = None
        self.spawner_class = None

    def create_derived_settings(self):
        '''Create further settings from passed-in ones.
        '''
        self.proxy_api_url = 'http://{}:{}'.format(
            self.proxy_host, self.proxy_api_port)
        self.bind_url = 'http://0.0.0.0:8000{}'.format(self.hub_route)
        self.hub_bind_url = 'http://0.0.0.0:8081{}'.format(self.hub_route)
        self.hub_connect_url = 'http://{}:{}{}'.format(self.hub_host,
                                                       self.hub_api_port,
                                                       self.hub_route)
        mm = self.lab_nodejs_max_mem
        self.lab_node_options = None
        if mm:
            self.lab_node_options = "--max-old-space-size={}".format(mm)
        while self.hub_route.endswith('/') and self.hub_route != '/':
            self.hub_route = self.hub_route[:-1]
        if not self.external_instance_url:
            ehu = self.external_hub_url
            if ehu:
                if ehu.endswith(self.hub_route):
                    self.external_instance_url = ehu[:-len(self.hub_route)]

    def dump(self):
        '''Return dict for pretty printing.
        '''
        myvars = vars(self)
        sanitized = sanitize_dict(myvars, ['reaper_password',
                                           'session_db_url'])
        # Stringify classrefs
        for key in ["log", "authenticator_class", "spawner_class"]:
            val = sanitized.get(key)
            if val:
                sanitized[key] = str(val)
        return sanitized

    def toJSON(self):
        return self.dump()



'''
JupyterHub Spawner to spawn user notebooks on a Kubernetes cluster in per-
user namespaces.

This module exports `NamespacedKubeSpawner` class, which is the spawner
implementation that should be used by JupyterHub.
'''

import sys
from eliot import start_action
from jupyterhub.utils import exponential_backoff
from kubernetes import client, config
from kubernetes.client.rest import ApiException
from kubernetes.config.config_exception import ConfigException
from kubespawner import KubeSpawner
from tornado import gen
from .multireflector import (MultiNamespacePodReflector,
                             MultiNamespaceEventReflector)
from ..utils import make_logger


class MultiNamespacedKubeSpawner(KubeSpawner):
    '''Implement a JupyterHub spawner to spawn pods in a Kubernetes Cluster
    with per-user namespaces.
    '''

    def __init__(self, *args, **kwargs):
        if not self.log:
            self.log = make_logger()
        super().__init__(*args, **kwargs)
        self.delete_grace_period = 25  # K8s takes ten or so, usually
        self.namespace = self.get_user_namespace()
        try:
            self.log.debug("Loading K8s config.")
            config.load_incluster_config()
        except ConfigException:
            config.load_kube_config()

    @gen.coroutine
    def poll(self):
        '''
        Check if the pod is still running.

        Uses the same interface as subprocess.Popen.poll(): if the pod is
        still running, returns None.  If the pod has exited, return the
        exit code if we can determine it, or 1 if it has exited but we
        don't know how.  These are the return values JupyterHub expects.

        Note that a clean exit will have an exit code of zero, so it is
        necessary to check that the returned value is None, rather than
        just Falsy, to determine that the pod is still running.
        '''
        # have to wait for first load of data before we have a valid answer
        with start_action(action_type="multispawner_poll"):
            if not self.pod_reflector.first_load_future.done():
                yield self.pod_reflector.first_load_future
            data = self.pod_reflector.pods.get((self.namespace, self.pod_name),
                                               None)
            if data is not None:
                if data.status.phase == 'Pending':
                    return None
                ctr_stat = data.status.container_statuses
                if ctr_stat is None:  # No status, no container (we hope)
                    # This seems to happen when a pod is idle-culled.
                    return 1
                for c in ctr_stat:
                    # return exit code if notebook container has terminated
                    if c.name == 'notebook':
                        if c.state.terminated:
                            # call self.stop to delete the pod
                            if self.delete_stopped_pods:
                                yield self.stop(now=True)
                            return c.state.terminated.exit_code
                        break
                # None means pod is running or starting up
                return None
            # pod doesn't exist or has been deleted
            return 1

    @gen.coroutine
    def _start(self):
        '''Start the user's pod.
        '''
        retry_times = 4  # Ad-hoc
        pod = yield self.get_pod_manifest()
        if self.modify_pod_hook:
            pod = yield gen.maybe_future(self.modify_pod_hook(self, pod))
        for i in range(retry_times):
            try:
                yield self.asynchronize(
                    self.api.create_namespaced_pod,
                    self.namespace,
                    pod,
                )
                break
            except ApiException as e:
                if e.status != 409:
                    # We only want to handle 409 conflict errors
                    self.log.exception("Failed for %s", pod.to_str())
                    raise
                self.log.info(
                    'Found existing pod %s, attempting to kill',
                    self.pod_name)
                # TODO: this should show up in events
                yield self.stop(now=True)

                self.log.info(
                    'Killed pod %s, will try starting ' % self.pod_name +
                    'singleuser pod again')
        else:
            raise Exception(
                'Can not create user pod %s :' % self.pod_name +
                'already exists and could not be deleted')

        # we need a timeout here even though start itself has a timeout
        # in order for this coroutine to finish at some point.
        # using the same start_timeout here
        # essentially ensures that this timeout should never propagate up
        # because the handler will have stopped waiting after
        # start_timeout, starting from a slightly earlier point.
        try:
            yield exponential_backoff(
                lambda: self.is_pod_running(self.pod_reflector.pods.get(
                    (self.namespace, self.pod_name), None)),
                'pod/%s did not start in %s seconds!' % (
                    self.pod_name, self.start_timeout),
                timeout=self.start_timeout,
            )
        except TimeoutError:
            if self.pod_name not in self.pod_reflector.pods:
                # if pod never showed up at all,
                # restart the pod reflector which may have
                # become disconnected.
                self.log.error(
                    "Pod {} never showed up in reflector; restarting " +
                    "pod reflector.".format(self.pod_name)
                )
                self._start_watching_pods(replace=True)
                raise

        pod = self.pod_reflector.pods[(self.namespace, self.pod_name)]
        self.pod_id = pod.metadata.uid
        return (pod.status.pod_ip, self.port)

    @gen.coroutine
    def stop(self, now=False):
        with start_action(action_type="stop"):
            delete_options = client.V1DeleteOptions()

            if now:
                grace_seconds = 0
            else:
                # Give it some time, but not the default (which is 30s!)
                grace_seconds = self.delete_grace_period

            delete_options.grace_period_seconds = grace_seconds
            self.log.info("Deleting pod %s", self.pod_name)
            try:
                yield self.asynchronize(
                    self.api.delete_namespaced_pod,
                    name=self.pod_name,
                    namespace=self.namespace,
                    body=delete_options,
                    grace_period_seconds=grace_seconds,
                )
            except ApiException as e:
                if e.status == 404:
                    self.log.warning(
                        "No pod %s to delete. Assuming already deleted.",
                        self.pod_name,
                    )
                else:
                    raise
            try:
                yield exponential_backoff(
                    lambda: self.pod_reflector.pods.get((self.namespace,
                                                         self.pod_name),
                                                        None) is
                    None,
                    'pod/%s did not disappear in %s seconds!' % (
                        self.pod_name, self.start_timeout),
                    timeout=self.start_timeout,
                )
            except TimeoutError:
                self.log.error(
                    "Pod %s did not disappear, " % self.pod_name +
                    "restarting pod reflector")
                self._start_watching_pods(replace=True)
                raise

    def _start_watching_events(self, replace=True):
        return self._start_reflector(
            "events",
            MultiNamespaceEventReflector,
            fields={"involvedObject.kind": "Pod"},
            replace=replace,
        )

    def _start_watching_pods(self, replace=True):
        return self._start_reflector("pods", MultiNamespacePodReflector,
                                     replace=replace)

    def _start_reflector(self, key, ReflectorClass, replace=True, **kwargs):

        def on_reflector_failure():
            self.log.critical(
                "%s reflector failed, halting Hub.",
                key.title(),
            )
            sys.exit(1)

        previous_reflector = self.__class__.reflectors.get(key)

        if replace or not previous_reflector:
            self.__class__.reflectors[key] = ReflectorClass(
                parent=self,
                namespace=self.namespace,
                on_failure=on_reflector_failure,
                **kwargs,
            )

        if replace and previous_reflector:
            # we replaced the reflector, stop the old one
            previous_reflector.stop()

        # return the current reflector
        return self.__class__.reflectors[key]



'''This is a mixin class for authenticators for common LSST functionality.
'''
import asyncio
import json
from eliot import start_action
from jupyterhub.auth import Authenticator
from .. import LSSTMiddleManager
from ..config import LSSTConfig


class LSSTAuthenticator(Authenticator):
    '''We create an LSST Manager structure on startup; this is an
    LSSTMiddleManager controlling a set of other managers: auth, env,
    namespace, quota, and volume.

    All LSSTAuthenticator subclasses are expected to create two new fields in
     auth_state:
      * auth_state['group_map'], which contains a dict mapping group name
        (the key) to a group ID number (the value).  GIDs may be strings or
        integers.
      * auth_state['uid'], which contains a string or an integer with the
        user's numeric UID.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lsst_mgr = LSSTMiddleManager(parent=self,
                                          authenticator=self,
                                          config=LSSTConfig())
        self.enable_auth_state = True
        self.delete_invalid_users = True
        self.token = None

    def dump(self):
        '''Return dict suitable for pretty-printing.
        '''
        ad = {"enable_auth_state": self.enable_auth_state,
              "delete_invalid_users": self.delete_invalid_users,
              "login_handler": str(self.login_handler),
              "lsst_mgr": self.lsst_mgr.dump()
              }
        return ad

    def toJSON(self):
        return json.dumps(self.dump())

    async def refresh_user(self, user, handler=None):
        '''On each refresh_user, clear the options form cache, thus
        forcing it to be rebuilt on next display.  Otherwise it is built once
        per user session, which is not frequent enough to display new images
        in a timely fashion.
        '''
        with start_action(action_type="refresh_user_lsstauth"):
            uname = user.escaped_name
            self.log.debug(
                "Entering lsstauth refresh_user for '{}'.".format(uname))
            self.log.debug("Calling superclass's refresh_user().")
            retval = await super().refresh_user(user, handler)
            self.log.debug("Returned from superclass's refresh_user().")
            self.log.debug("Clearing form data for '{}'.".format(uname))
            self.lsst_mgr.optionsform_mgr.options_form_data = None
            self.log.debug("Adding uid and groupmap to auth_state.")
            self.log.debug(
                "Finished lsstauth refresh_user for '{}'".format(uname))
            return retval



'''Classes and tool for scanning Docker repositories.
'''
from .scanrepo import ScanRepo
from .standalone import standalone
from .singletonscanner import SingletonScanner
from .reaper import Reaper
from .reaperstandalone import reaperstandalone
from .prepuller import Prepuller
from .prepullerstandalone import prepullerstandalone
from .parse_args import parse_args
from .primerepocache import prime_repo_cache
__all__ = [ScanRepo, SingletonScanner, Reaper, Prepuller,
           standalone, reaperstandalone, prepullerstandalone, parse_args,
           prime_repo_cache]



'''Class to manage an LSST-specific options form.
'''
import datetime
import jinja2
import json
from collections import OrderedDict
from eliot import start_action
from time import sleep
from .. import SingletonScanner as SScan
from .. import LoggableChild


class LSSTOptionsFormManager(LoggableChild):
    '''Class to create and read a spawner form.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sizemap = {}
        self._scanner = None
        self.options_form_data = None

    def get_options_form(self):
        '''Create an LSST Options Form from parent's config object.
        '''
        with start_action(action_type="get_options_form"):
            # Make options form by scanning container repository, then cache.
            # For a single spawning session, you always get the same form.
            #
            # If that's not OK (long-lived tokens, for example) then in
            #  your authenticator's refresh_user(), clear options_form_data.
            uname = self.parent.user.escaped_name
            self.log.debug("Creating options form for '{}'".format(uname))
            if self.options_form_data:
                self.log.debug("Returning cached options form.")
                return self.options_form_data
            self.log.debug("Regenerating form_data for '{}'.".format(uname))
            cfg = self.parent.config
            scanner = SScan(host=cfg.lab_repo_host,
                            owner=cfg.lab_repo_owner,
                            name=cfg.lab_repo_name,
                            experimentals=cfg.prepuller_experimentals,
                            dailies=cfg.prepuller_dailies,
                            weeklies=cfg.prepuller_weeklies,
                            releases=cfg.prepuller_releases,
                            cachefile=cfg.prepuller_cachefile,
                            debug=cfg.debug)
            self._scanner = scanner
            self.log.debug("Calling _sync_scan() for '{}'.".format(uname))
            self._sync_scan()
            self.log.debug("Back from _sync_scan() for '{}'.".format(uname))
            lnames, ldescs = scanner.extract_image_info()
            desclist = []
            # Setting this up to pass into the Jinja template more easily
            for idx, img in enumerate(lnames):
                desclist.append({"name": img,
                                 "desc": ldescs[idx]})
            colon = lnames[0].find(':')
            custtag = lnames[0][:colon] + ":__custom"
            all_tags = scanner.get_all_tags()
            now = datetime.datetime.now()
            nowstr = now.ctime()
            if not now.tzinfo:
                # If we don't have tzinfo, assume it's in UTC
                nowstr += " UTC"
            self.log.debug(
                "About to call _make_sizemap() for '{}'.".format(uname))
            self._make_sizemap()
            self.log.debug(
                "Back from _make_sizemap() for '{}'.".format(uname))
            # in order to get the right default size index, we need to poke the
            #  quota manager, because different users may get different sizes
            #  by default
            cfg = self.parent.config
            qm = self.parent.quota_mgr
            qm.define_resource_quota_spec()
            defaultsize = (qm.custom_resources.get('size_index') or
                           cfg.size_index)
            template_file = self.parent.config.form_template
            template_loader = jinja2.FileSystemLoader(searchpath='/')
            template_environment = jinja2.Environment(loader=template_loader)
            template = template_environment.get_template(template_file)
            optform = template.render(
                defaultsize=defaultsize,
                desclist=desclist,
                all_tags=all_tags,
                custtag=custtag,
                sizelist=list(self.sizemap.values()),
                nowstr=nowstr)
            self.options_form_data = optform
            self.log.debug(
                "Generated options_form_data for '{}'.".format(uname))
            return optform

    def resolve_tag(self, tag):
        '''Delegate to scanner to resolve convenience tags.
        '''
        with start_action(action_type="resolve_tag"):
            self.log.debug("Resolving tag for '{}'.".format(tag))
            rtag = self._scanner.resolve_tag(tag)
            self.log.debug("Resolved tag for '{}'->'{}'.".format(tag, rtag))
            return rtag

    def _sync_scan(self):
        with start_action(action_type="_sync_scan"):
            uname = self.parent.user.escaped_name
            self.log.debug("Entering _sync_scan() for '{}'.".format(uname))
            scanner = self._scanner
            cfg = self.parent.config
            delay_interval = cfg.initial_scan_interval
            max_delay_interval = cfg.max_scan_interval
            max_delay = cfg.max_scan_delay
            delay = 0
            epoch = datetime.datetime(1970, 1, 1)
            while scanner.last_updated == epoch:
                self.log.info(("Scan results not available yet; sleeping " +
                               "{:02.1f}s ({:02.1f}s " +
                               "so far).").format(delay_interval, delay))
                sleep(delay_interval)
                delay = delay + delay_interval
                delay_interval *= 2
                if delay_interval > max_delay_interval:
                    delay_interval = max_delay_interval
                if delay >= max_delay:
                    errstr = ("Scan results did not become available in " +
                              "{}s.".format(max_delay))
                    raise RuntimeError(errstr)
            self.log.debug("Completed _sync_scan() for '{}'.".format(uname))

    def _make_sizemap(self):
        with start_action(action_type="_make_sizemap"):
            uname = self.parent.user.escaped_name
            self.log.debug("Entering _make_sizemap() for '{}'.".format(uname))
            sizemap = OrderedDict()
            # For supported Python versions, dict is ordered anyway...
            sizes = self.parent.config.form_sizelist
            tiny_cpu = self.parent.config.tiny_cpu_max
            mb_per_cpu = self.parent.config.mb_per_cpu
            size_range = float(self.parent.config.lab_size_range)
            # Each size doubles the previous one.
            cpu = tiny_cpu
            idx = 0
            for sz in sizes:
                mem_mb = int(mb_per_cpu * cpu)
                sizemap[sz] = {"cpu": cpu,
                               "mem": "{}M".format(mem_mb),
                               "name": sz,
                               "index": idx}
                min_cpu = float(cpu / size_range)
                sizemap[sz]["cpu_guar"] = min_cpu
                min_mem_mb = int(mb_per_cpu * min_cpu)
                sizemap[sz]["mem_guar"] = "{}M".format(min_mem_mb)
                desc = sz.title() + " (%.2f CPU, %dM RAM)" % (cpu, mem_mb)
                sizemap[sz]["desc"] = desc
                cpu = cpu * 2
                idx = idx + 1
            # Clean up if list of sizes changed.
            self.sizemap = sizemap
            self.log.debug("Finished _make_sizemap() for '{}'.".format(uname))

    def options_from_form(self, formdata=None):
        '''Get user selections.
        '''
        with start_action(action_type="options_from_form"):
            options = None
            if formdata:
                self.log.debug("Form data: %s", json.dumps(formdata,
                                                           sort_keys=True,
                                                           indent=4))
                options = {}
                if ('kernel_image' in formdata and formdata['kernel_image']):
                    options['kernel_image'] = formdata['kernel_image'][0]
                if ('size' in formdata and formdata['size']):
                    options['size'] = formdata['size'][0]
                if ('image_tag' in formdata and formdata['image_tag']):
                    options['image_tag'] = formdata['image_tag'][0]
                if ('clear_dotlocal' in formdata and
                        formdata['clear_dotlocal']):
                    options['clear_dotlocal'] = True
                if ('enable_debug' in formdata and
                        formdata['enable_debug']):
                    options['enable_debug'] = True
            return options

    def dump(self):
        '''Return contents dict for pretty-printing.
        '''
        sd = {"parent": str(self.parent),
              "sizemap": self.sizemap,
              "scanner": str(self._scanner)}
        return sd

    def toJSON(self):
        return json.dumps(self.dump())



'''LSST Authentication classes.
'''
from .lsstauth import LSSTAuthenticator
from .lsstjwtauth import LSSTJWTAuthenticator
from .lsstjwtloginhandler import LSSTJWTLoginHandler

__all__ = [LSSTAuthenticator, LSSTJWTAuthenticator, LSSTJWTLoginHandler]



import json
from eliot import start_action
from .. import Loggable
from .apimanager import LSSTAPIManager
from .envmanager import LSSTEnvironmentManager
from .namespacemanager import LSSTNamespaceManager
from .optionsformmanager import LSSTOptionsFormManager
from .quotamanager import LSSTQuotaManager
from .volumemanager import LSSTVolumeManager


class LSSTMiddleManager(Loggable):
    '''The LSSTMiddleManager is a class that holds references to various
    LSST-specific management objects and delegates requests to them.
    The idea is that an LSST Spawner could instantiate a single
    LSSTMiddleManager, which would then be empowered to perform all
    LSST-specific operations, reducing configuration complexity.

    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = kwargs.pop('config', None)
        self.parent = kwargs.pop('parent', None)
        self.log.info(
            "Parent of LSST Middle Manager is '{}'".format(self.parent))
        self.authenticator = kwargs.pop('authenticator', None)
        self.spawner = kwargs.pop('spawner', None)
        self.user = kwargs.pop('user', None)
        self.api_mgr = LSSTAPIManager(parent=self)
        self.env_mgr = LSSTEnvironmentManager(parent=self)
        self.namespace_mgr = LSSTNamespaceManager(parent=self)
        self.optionsform_mgr = LSSTOptionsFormManager(parent=self)
        self.quota_mgr = LSSTQuotaManager(parent=self)
        self.volume_mgr = LSSTVolumeManager(parent=self)
        self.api = self.api_mgr.api
        self.rbac_api = self.api_mgr.rbac_api

    def ensure_resources(self):
        '''Delegate to namespace manager.
        '''
        with start_action(action_type="ensure_resources"):
            self.namespace_mgr.ensure_namespace()

    def dump(self):
        '''Return contents dict to pretty-print.
        '''
        md = {"parent": str(self.parent),
              "authenticator": str(self.authenticator),
              "api": str(self.api),
              "rbac_api": str(self.rbac_api),
              "config": self.config.dump(),
              "api_mgr": self.api_mgr.dump(),
              "env_mgr": self.env_mgr.dump(),
              "optionsform_mgr": self.optionsform_mgr.dump(),
              "quota_mgr": self.quota_mgr.dump(),
              "volume_mgr": self.volume_mgr.dump()
              }
        if self.user:
            md["user"] = "{}".format(self.user)
        if self.spawner:
            md["spawner"] = self.spawner.dump()
        return md

    def toJSON(self):
        return json.dumps(self.dump())



'''
Shared utility functions.
'''

import hashlib
import inspect
import logging
import os
import requests

from collections import defaultdict
from eliot.stdlib import EliotHandler


def rreplace(s, old, new, occurrence):
    '''Convenience function from:
    https://stackoverflow.com/questions/2556108/\
    rreplace-how-to-replace-the-last-occurrence-of-an-expression-in-a-string
    '''
    li = s.rsplit(old, occurrence)
    return new.join(li)


def sanitize_dict(input_dict, sensitive_fields):
    '''Remove sensitive content.  Useful for logging.
    '''
    retval = {}
    if not input_dict:
        return retval
    retval.update(input_dict)
    for field in sensitive_fields:
        if retval.get(field):
            retval[field] = "[redacted]"
    return retval


def get_execution_namespace():
    '''Return Kubernetes namespace of this container.
    '''
    ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
    if os.path.exists(ns_path):
        with open(ns_path) as f:
            return f.read().strip()
    return None


def make_logger(name=None, level=None):
    '''Create a logger with LSST-appropriate characteristics.
    '''
    if name is None:
        # Get name of caller's class.
        #  From https://stackoverflow.com/questions/17065086/
        frame = inspect.stack()[1][0]
        name = _get_classname_from_frame(frame)
    logger = logging.getLogger(name)
    if name is None:
        logger.info('jupyterhubutils make_logger() called for root logger.')
        logger.info('not eliotify-ing root logger.')
        return logger
    logger.propagate = False
    if level is None:
        level = logging.getLogger().getEffectiveLevel()
    logger.setLevel(level)
    logger.handlers = [EliotHandler()]
    logger.info("Created logger object for class '{}'.".format(name))
    return logger


def _get_classname_from_frame(fr):
    args, _, _, value_dict = inspect.getargvalues(fr)
    # we check the first parameter for the frame function is
    # named 'self'
    if len(args) and args[0] == 'self':
        # in that case, 'self' will be referenced in value_dict
        instance = value_dict.get('self', None)
        if instance:
            # return its classname
            cl = getattr(instance, '__class__', None)
            if cl:
                return "{}.{}".format(cl.__module__, cl.__name__)
    # If it wasn't a class....
    return '<unknown>'


def str_bool(s):
    '''Make a sane guess for whether a value represents true or false.
    Intended for strings, mostly in the context of environment variables,
    but if you pass it something that's not a string that is falsy, like
    an empty list, it will cheerfully return False.
    '''
    if not s:
        return False
    if type(s) != str:
        # It's not a string and it's not falsy, soooo....
        return True
    s = s.lower()
    if s in ['false', '0', 'no', 'n']:
        return False
    return True


def str_true(v):
    '''The string representation of a true value will be 'TRUE'.  False will
    be the empty string.
    '''
    if v:
        return 'TRUE'
    else:
        return ''


def listify(item, delimiter=','):
    '''Used for taking character (usually comma)-separated string lists
    and returning an actual list, or the empty list.
    Useful for environment parsing.

    Sure, you could pass it integer zero and get [] back.  Don't.
    '''
    if not item:
        return []
    if type(item) is str:
        item = item.split(delimiter)
    if type(item) is not list:
        raise TypeError("'listify' must take None, str, or list!")
    return item


def floatify(item, default=0.0):
    '''Another environment-parser: the empty string should be treated as
    None, and return the default, rather than the empty string (which
    does not become an integer).  Default can be either a float or string
    that float() works on.  Note that numeric zero (or string '0') returns
    0.0, not the default.  This is intentional.
    '''
    if item is None:
        return default
    if item == '':
        return default
    return float(item)


def intify(item, default=0):
    '''floatify, but for ints.
    '''
    return int(floatify(item, default))


def list_duplicates(seq):
    '''List duplicate items from a sequence.
    '''
    # https://stackoverflow.com/questions/5419204
    tally = defaultdict(list)
    for i, item in enumerate(seq):
        tally[item].append(i)
    return ((key, locs) for key, locs in tally.items()
            if len(locs) > 1)


def list_digest(inp_list):
    '''Return a digest to uniquely identify a list.
    '''
    if type(inp_list) is not list:
        raise TypeError("list_digest only works on lists!")
    if not inp_list:
        raise ValueError("input must be a non-empty list!")
    # If we can rely on python >= 3.8, shlex.join is better
    return hashlib.sha256(' '.join(inp_list).encode('utf-8')).hexdigest()


def get_access_token(tokenfile=None):
    '''Determine the access token from the mounted secret or environment.
    '''
    tok = None
    hdir = os.environ.get('HOME', None)
    if hdir:
        if not tokenfile:
            # FIXME we should make this instance-dependent
            tokfile = hdir + "/.access_token"
        try:
            with open(tokfile, 'r') as f:
                tok = f.read().replace('\n', '')
        except Exception as exc:
            log = make_logger()
            log.warn("Could not read tokenfile '{}': {}".format(tokfile, exc))
    if not tok:
        tok = os.environ.get('ACCESS_TOKEN', None)
    return tok


def parse_access_token(endpoint=None, tokenfile=None, token=None, timeout=15):
    '''Rely on gafaelfawr to validate and parse the access token.
    '''
    if not token:
        token = get_access_token(tokenfile=tokenfile)
    if not token:
        raise RuntimeError("Cannot determine access token!")
    # Endpoint is constant in an ArgoCD-deployed cluster
    if not endpoint:
        endpoint = "http://gafaelfawr-service.gafaelfawr:8080/auth/analyze"
    resp = requests.post(endpoint, data={'token': token}, timeout=timeout)
    rj = resp.json()
    p_resp = rj["token"]
    if not p_resp["valid"]:
        raise RuntimeError("Access token invalid: '{}'!".format(str(resp)))
    # Force to lowercase username (should no longer be necessary)
    p_tok = p_resp["data"]
    uname = p_tok["uid"]
    p_tok["uid"] = uname.lower()
    return p_tok


def get_fake_gid(grpname):
    '''Use if we have strict_ldap_groups off, to assign GIDs to names
    with no matching Unix GID.  We would like them to be consistent, so
    we will use a hash of the group name, modulo some large-ish constant,
    added to another large-ish constant.

    There is a chance of collision, but it doesn't really matter.

    We do need to keep the no-GID groups around, though, because we might
    be using them to make options form or quota decisions (if we know we
    don't, we should turn on strict_ldap_groups).
    '''
    grpbase = 3E7
    grprange = 1E7
    grphash = hashlib.sha256(grpname.encode('utf-8')).hexdigest()
    grpint = int(grphash, 16)
    igrp = int(grpbase + (grpint % grprange))
    return igrp


def make_passwd_line(claims):
    '''Create an entry for /etc/passwd based on our claims.  Returns a
    newline-terminated string.
    '''
    uname = claims['uid']
    uid = claims['uidNumber']
    pwline = "{}:x:{}:{}::/home/{}:/bin/bash\n".format(
        uname, uid, uid, uname)
    return pwline


def assemble_gids(claims, strict_ldap=False):
    '''Take the claims data and return the string to be used for be used
    for provisioning the user and groups (in sudo mode).
    '''
    glist = _map_supplemental_gids(claims, strict_ldap=strict_ldap)
    gidlist = ["{}:{}".format(x[0], x[1]) for x in glist]
    return ','.join(gidlist)


def make_group_lines(claims, strict_ldap=False):
    '''Create a list of newline-terminated strings representing group
    entries suitable for appending to /etc/group.
    '''
    uname = claims['uid']
    uid = claims['uidNumber']
    # Add individual group; don't put user in it (implicit from group in
    #  passwd)
    glines = ["{}:x:{}:\n".format(uname, uid)]
    glist = _map_supplemental_gids(claims, strict_ldap=strict_ldap)
    glines.extend(["{}:x:{}:{}\n".format(x[0], x[1], uname) for x in glist])
    return glines


def get_supplemental_gids(claims, strict_ldap=False):
    '''Create a list of gids suitable to paste into the supplemental_gids
    the container can run with (in sudoless mode).'''
    glist = _map_supplemental_gids(claims, strict_ldap=strict_ldap)
    return [x[1] for x in glist]


def resolve_groups(claims, strict_ldap=False):
    '''Returns groupmap suitable for insertion into auth_state;
    group values are strings.
    '''
    glist = _map_supplemental_gids(claims, strict_ldap=strict_ldap)
    groupmap = {}
    for gt in glist:
        groupmap[gt[0]] = str(gt[1])
    return groupmap


def _map_supplemental_gids(claims, strict_ldap=False):
    '''Helper function to deal with group manipulation.  Returns a list of
    tuples (groupname, gid).

    If a name has no id, omit the entry if strict_ldap is True.  Otherwise
    generate a fake gid for it and use that.
    '''
    uname = claims['uid']
    groups = claims['isMemberOf']
    retval = []
    for grp in groups:
        gname = grp['name']
        if gname == uname:
            continue  # We already have private group as runAsGid
        gid = grp.get('id', None)
        if not gid:
            if not strict_ldap:
                gid = get_fake_gid(gname)
        if gid:
            retval.append((gname, gid))
    return retval


def add_user_to_groups(uname, grpstr, groups=['lsst_lcl', 'jovyan']):
    '''Take a user name (a string) and a base group file (as a string) and
    inject the user into the appropriate groups, given in the groups
    parameter (defaults to 'lsst_lcl' and 'jovyan').  Returns a string.'''
    glines = grpstr.split('\n')
    g_str = ''
    for grp in glines:
        s_line = grp.strip()
        if not s_line:
            continue
        grpname = s_line.split(':')[0]
        if grpname in groups:
            if s_line.endswith(':'):
                s_line = s_line + uname
            else:
                s_line = s_line + ',' + uname
        g_str = g_str + s_line + '\n'
    return g_str



'''LSST Login Handler to use JWT token present in request headers.
'''
from eliot import start_action
from tornado import gen, web
from jupyterhub.utils import url_path_join
from jwtauthenticator.jwtauthenticator import JSONWebTokenLoginHandler
from ..utils import parse_access_token


class LSSTJWTLoginHandler(JSONWebTokenLoginHandler):

    @gen.coroutine
    def get(self):
        '''Authenticate on get() via reading the token from HTTP headers.
        '''
        # This causes some token context error in Eliot
        # with start_action(action_type="get"):
        _ = yield self._jwt_authenticate()
        #
        # We can't just use the superclass because it issues the redirect,
        #  and we can't do that until we do our additional checking.
        _url = url_path_join(self.hub.server.base_url, 'home')
        next_url = self.get_argument('next', default=False)
        if next_url:
            _url = next_url
        self.redirect(_url)

    @gen.coroutine
    def post(self):
        '''Also authenticate on POST, if necessary.'''
        with start_action(action_type="post"):
            _ = yield(self._jwt_authenticate())
            yield super().post()

    @gen.coroutine
    def _jwt_authenticate(self):
        # This is taken from https://github.com/mogthesprog/jwtauthenticator
        #  but with our additional claim information checked and stuffed
        #  into auth_state.
        # Gafaelfawr will not give us back an invalid token; we will raise an
        #  error if something goes wrong with the token analysis
        claims, token = yield self._check_auth_header()
        username_claim_field = self.authenticator.username_claim_field
        username = self.retrieve_username(claims, username_claim_field).lower()
        user = self.user_from_username(username)
        # Here is where we deviate from the vanilla JWT authenticator.
        # We simply store all the JWT claims in auth_state, although we
        #  choose our field names to make the spawner reusable from the
        #  OAuthenticator implementation.
        auth_state = yield self.refresh_user(user)
        _ = yield user.save_auth_state(auth_state)
        self.set_login_cookie(user)

    @gen.coroutine
    def refresh_user(self, user, handler=None):
        '''Validate the token and force re-auth if the claims are not
        (presumably no longer) valid.
        '''
        with start_action(action_type="refresh_user_loginhandler"):
            uname = user.escaped_name
            self.log.debug(
                "Entering refresh_user() on jwtloginhandler for '{}'.".format(
                    uname))
            try:
                self.log.debug("Checking auth header for '{}'.".format(uname))
                claims, token = yield self._check_auth_header()
                self.log.debug("Auth header for '{}' is valid.".format(uname))
            except web.HTTPError:
                # Force re-login
                self.log.debug(
                    "Auth header check for '{}' failed.".format(uname))
                return False
            username_claim_field = self.authenticator.username_claim_field
            username = self.retrieve_username(
                claims, username_claim_field).lower()
            auth_state = {"id": username,
                          "access_token": token,
                          "claims": claims}
            self.log.debug(
                "Finished jwtloginhandler refresh_user for '{}'.".format(
                    uname))
            return auth_state

    @gen.coroutine
    def _check_auth_header(self):
        with start_action(action_type="_check_auth_header"):
            # Either returns (valid) claims and token,
            #  or throws a web error of some type.
            self.log.debug("Checking authentication header.")
            auth = self.authenticator
            header_name = auth.header_name
            param_name = auth.param_name
            header_is_authorization = auth.header_is_authorization
            auth_header_content = self.request.headers.get(header_name, "")
            auth_cookie_content = self.get_cookie("XSRF-TOKEN", "")
            tokenParam = self.get_argument(param_name, default=False)
            if auth_header_content and tokenParam:
                self.log.error("Authentication: both an authentication " +
                               "header and tokenParam")
                raise web.HTTPError(400)
            elif auth_header_content:
                if header_is_authorization:
                    # We should not see "token" as first word in the
                    #  authorization header.  If we do it could mean someone
                    #  coming in with a stale API token
                    if auth_header_content.split()[0].lower() != "bearer":
                        self.log.error("Authorization header is not 'bearer'.")
                        raise web.HTTPError(403)
                    token = auth_header_content.split()[1]
                else:
                    token = auth_header_content
            elif auth_cookie_content:
                token = auth_cookie_content
            elif tokenParam:
                token = tokenParam
            else:
                self.log.error("Could not determine authentication token.")
                raise web.HTTPError(401)
            # Delegate to gafaelfawr for validation.
            self.log.debug("Sending token to gafaelfawr for validation.")
            try:
                claims = parse_access_token(token=token)
            except RuntimeError as exc:
                self.log.error("Token validation failed: '{}'".format(exc))
                raise web.HTTPError(403)
            uname = claims['uid']
            self.log.debug(
                "Gafaelfawr validated token for '{}'.".format(uname))
            return claims, token
